#lang racket
(require parser-tools/lex
         parser-tools/yacc)
(require "declarations.rkt")
(require "utilities.rkt")


;(define (buildNullable t)
 ; ...)

;(define (buildFirst t)
 ; ...)

;(define (buildLast t)
 ; ...)

;(define (buildFollow t)
 ; ...)

;(define (buildGraph reg)
;  ...)

(define (buildNullable t)
  (match t
    [(Epsilon n) (list (cons n #t))]
    [(Literal c n) (list (cons n #f))]
    [(Or t1 t2 n) (let* [(x1 (buildNullable t1))
                         (x2 (buildNullable t2))]
                    
                    (cons (cons n (or (cdar x1) (cdar x2))) (append x1 x2)))]
    [(Then t1 t2 n) (let* [(x1 (buildNullable t1))
                           (x2 (buildNullable t2))]
                      (cons (cons n (and (cdar x1) (cdar x2))) (append x1 x2)))]
    [(Star t n) (cons (cons n #t) (buildNullable t))]))

(define (buildFirst t)
  (match t
    [(Epsilon n) (list (list n))]
    [(Literal c n) (list (list n n))]
    [(Or t1 t2 n) (let* ([y1 (buildFirst t1)]
                         [y2 (buildFirst t2)])
                    (cons (cons n (append (cdar y1) (cdar y2))) (append y1 y2)))]
    [(Then t1 t2 n) (let* [(x1 (buildNullable t1))
                           (y1 (buildFirst t1))
                           (y2 (buildFirst t2))]
                      (cond [(cdar x1) (cons (cons n (append (cdar y1) (cdar y2))) (append y1 y2))]
                            [else (cons (cons n (cdar y1)) (append y1 y2))]))]
    [(Star t n) (let* ([y (buildFirst t)])
                  (cons (cons n (cdar y)) y))]))

(define (buildLast t)
  (match t
    [(Epsilon n) (list (list n))]
    [(Literal c n) (list (list n n))]
    [(Or t1 t2 n) (let* ([y1 (buildLast t1)]
                         [y2 (buildLast t2)])
                    (cons (cons n (append (cdar y1) (cdar y2))) (append y1 y2)))]
    [(Then t1 t2 n) (let* [(x2 (buildNullable t2))
                           (y1 (buildLast t1))
                           (y2 (buildLast t2))]
                      (cond [(cdar x2) (cons (cons n (append (cdar y1) (cdar y2))) (append y1 y2))]
                            [else (cons (cons n (cdar y2)) (append y1 y2))]))]
    [(Star t n) (let* ([y (buildLast t)])
                  (cons (cons n (cdar y)) y))]))

(define (buildFollow t)
  (match t
    [(Literal c n)  (if (equal? "#" c) '() (list (list n)))]
    [(Epsilon n) '()]
    [(Or t1 t2 n) (append (buildFollow t1) (buildFollow t2))]
    [(Then t1 t2 n) (let* [(z1 (cdar (buildLast t1)))
                           (y2 (cdar (buildFirst t2)))
                           (w1 (buildFollow t1))
                           (w2 (buildFollow t2))
                           (w3 (map (lambda (l1) (cond [(member (car l1) z1) (append l1 y2) ]
                                                       [else l1])) w1))]
                      (append w3 w2))]
    [(Star t n) (let* [(y (cdar (buildFirst t)))
                       (z (cdar (buildLast t)))
                       (w (buildFollow t))]
                  (map (lambda (l1) (cond [(member (car l1) z) (cons (car l1) (set-union (cdr l1) y))]
                                          [else l1])) w))]))
(define (zip l1 l2)
  (cond [(or (null? l1) (null? l2)) '()]
        [else (cons (cons (car l1)(car l2)) (zip (cdr l1)(cdr l2)))]))
(define (my-last l)
  [cond [(null? l) "error"]
        [(null? (cdr l)) (car l)]
        [else (my-last (cdr l))]])

(define (buildGraph r)
  (let* ([t (maketree r)])
    (define (give-literals t)
      (define (give-h t p)
        (match t
          [(Literal c n) (cons (cons c (car p)) (cons (cons c n) (cdr p)))]
          [(Or t1 t2 n) (let* ([x1 (give-literals t1)]
                               [x2 (give-literals t2)])
                          (cons (append (car x1) (car x2)) (append (cdr x1) (cdr x2))))]
          [(Then t1 t2 n) (let* ([x1 (give-literals t1)]
                                 [x2 (give-literals t2)])
                            (cons (append (car x1) (car x2)) (append (cdr x1) (cdr x2))))]
          [(Star t n) (give-literals t)]
          [(Epsilon n)  (cons (cons "@" (car p)) (cons (cons "@" n) (cdr p)))]))
      (give-h t (cons '() '())))
    (define (remove-repeat p l)r
      (cond[(null? (cdr p)) l]
           [else (cond[(member (car p) l) (remove-repeat (cdr p) l)]
                      [else (remove-repeat (cdr p) (cons (car p) l))])]))
    (define (give-no-r t)  
      (let* ([v (give-literals t)]
             [v1 (car v)]
             [v2 (cdr v)])
        (cons (remove-repeat v1 '()) v2)))
    (define (green-node t)
      (cdar (buildFirst t)))
    (define (give-big lol)
      (if (null? (cdr lol)) (cdar lol) (give-big (cdr lol))))
    
    (let* ([gl (give-no-r t)]
           [bf1 (buildFollow t)]
           [gl1 (car gl)]
           [gl2 (cdr gl)]
           [big-literal (give-big gl2)])   
      
      (define (trans l x)
        (define (trans-h n bf)
          (if (= n (caar bf)) (cdar bf) (trans-h n (cdr bf))))
        (let* ([ lngth (length l)]
               [ mklst (make-list lngth x)]
               [zip-l-x (zip mklst l)]
               [set-inter (set-intersect zip-l-x gl2)]
               [possible (map (lambda (p) (cdr p)) set-inter)]
               [new-node (sort (remove-duplicates (append* (map (lambda (n) (trans-h n bf1)) possible))) <)])
          (cons new-node (Trans l x new-node))))
      
      (define (give-node-translist god)
        (if (null? god) '() 
            (let* ([caar (caar god)])
              (if (null? caar ) (give-node-translist (cdr god)) (cons (cons caar (cdar god)) (give-node-translist (cdr god)))))))
      
      (define (trans-node total-n new-n total-t new-t)
        (if (and (null? new-n) (null? new-t)) (cons total-n total-t) (let* ([god-list (append* (map(lambda (l) (map (lambda (x) (trans l x)) gl1)) new-n))]
                                                                            [node-trans-list (give-node-translist god-list)]
                                                                            [trans-list (map (lambda (p) (cdr p)) node-trans-list)]
                                                                            [node-list (map (lambda (p) (car p)) node-trans-list)]
                                                                            [new-nodes (set-subtract node-list total-n)]
                                                                            [new-total-nodes (set-union new-nodes total-n)]
                                                                            [new-trans (set-subtract trans-list new-t)]
                                                                            [new-total-trans (set-union new-trans total-t)])
                                                                       (trans-node new-total-nodes new-nodes new-total-trans new-trans))))
      (let* ([greennode (green-node t)]
             [trans-node-1 (trans-node (list greennode) (list greennode) '() '())]
             [all-nodes (car trans-node-1)]
             [all-trans (cdr trans-node-1)]
             [rednodes (filter (lambda (lx) (if (= big-literal (my-last lx)) #t #f)) all-nodes)])
        (Graph greennode all-nodes all-trans rednodes (append (remove "@" gl1) (list "#")))))))
